/*------------------------------------------------------------------------
 *  Copyright 2007-2009 (c) Jeff Brown <spadix@users.sourceforge.net>
 *
 *  This file is part of the ZBar Bar Code Reader.
 *
 *  The ZBar Bar Code Reader is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU Lesser Public License as
 *  published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *
 *  The ZBar Bar Code Reader is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser Public License
 *  along with the ZBar Bar Code Reader; if not, write to the Free
 *  Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 *  Boston, MA  02110-1301  USA
 *
 *  http://sourceforge.net/projects/zbar
 *------------------------------------------------------------------------*/

#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#ifdef _WIN32
# include <io.h>
# include <fcntl.h>
#endif
#include <assert.h>

#include <zbar.h>

#define BELL "\a"

static const char *note_usage =
    "usage: zbarcam [options] [/dev/video?]\n"
    "\n"
    "scan and decode bar codes from a video stream\n"
    "\n"
    "options:\n"
    "    -h, --help      display this help text\n"
    "    --version       display version information and exit\n"
    "    -q, --quiet     disable beep when symbol is decoded\n"
    "    -v, --verbose   increase debug output level\n"
    "    --verbose=N     set specific debug output level\n"
    "    --xml           use XML output format\n"
    "    --raw           output decoded symbol data without symbology prefix\n"
    "    --TGXf          decode symbol data via the Thru|Glass|Xfer protocol\n"
    "    --nodisplay     disable video display window\n"
    "    --prescale=<W>x<H>\n"
    "                    request alternate video image size from driver\n"
    "    -S<CONFIG>[=<VALUE>], --set <CONFIG>[=<VALUE>]\n"
    "                    set decoder/scanner <CONFIG> to <VALUE> (or 1)\n"
    /* FIXME overlay level */
    "\n";

static const char *xml_head =
    "<barcodes xmlns='http://zbar.sourceforge.net/2008/barcode'>"
    "<source device='%s'>\n";
static const char *xml_foot =
    "</source></barcodes>\n";

static zbar_processor_t *proc;
static int quiet = 0;
static enum {
    DEFAULT, RAW, TGXF, XML
} format = DEFAULT;

static char *xml_buf = NULL;
static unsigned xml_len = 0;

static int usage (int rc)
{
    FILE *out = (rc) ? stderr : stdout;
    fprintf(out, "%s", note_usage);
    return(rc);
}

static inline int parse_config (const char *cfgstr, int i, int n, char *arg)
{
    if(i >= n || !*cfgstr) {
        fprintf(stderr, "ERROR: need argument for option: %s\n", arg);
        return(1);
    }
    
    if(zbar_processor_parse_config(proc, cfgstr)) {
        fprintf(stderr, "ERROR: invalid configuration setting: %s\n", cfgstr);
        return(1);
    }
    return(0);
}




// The following CRC code comes from RFC1952, Section 8
//   http://tools.ietf.org/html/rfc1952#section-8

/* Table of CRCs of all 8-bit messages. */
unsigned long crc_table[256];

/* Flag: has the table been computed? Initially false. */
int crc_table_computed = 0;

/* Make the table for a fast CRC. */
void make_crc_table(void)
{
  unsigned long c;
  int n, k;
  for (n = 0; n < 256; n++) {
    c = (unsigned long) n;
    for (k = 0; k < 8; k++) {
      if (c & 1) {
        c = 0xedb88320L ^ (c >> 1);
      } else {
        c = c >> 1;
      }
    }
    crc_table[n] = c;
  }
  crc_table_computed = 1;
}


/*
    Update a running crc with the bytes buf[0..len-1] and return
  the updated crc. The crc should be initialized to zero. Pre- and
  post-conditioning (one's complement) is performed within this
  function so it shouldn't be done by the caller. Usage example:

  unsigned long crc = 0L;

  while (read_buffer(buffer, length) != EOF) {
    crc = update_crc(crc, buffer, length);
  }
  if (crc != original_crc) error();
*/
unsigned long update_crc(unsigned long crc,
  unsigned char *buf, int len)
{
  unsigned long c = crc ^ 0xffffffffL;
  int n;

  if (!crc_table_computed)
    make_crc_table();
  for (n = 0; n < len; n++) {
    c = crc_table[(c ^ buf[n]) & 0xff] ^ (c >> 8);
  }
  return c ^ 0xffffffffL;
}

/* Return the CRC of the bytes buf[0..len-1]. */
unsigned long crc(unsigned char *buf, int len)
{
  return update_crc(0L, buf, len);
}




/*

                                   _JNJ`
                               .JNMH`
                              JMMF`       `;.
                            .NMM)          `MN.
                            MMM)            (MML
                           (MMM`             MMML
                     M  I  D  N  I  G  H  T    C  o  D  E
                           (NMMF             MHNH
                            NMML            .MMM
                             NMML          .NMH
                              4MMNL       .#F
                               `4HNNL_   `
                                   `"""`

                          Copyright (C) 2004-2014
       "Ian (Larry) Latter" <ian dot latter at midnightcode dot org>

    Midnight Code is a registered trademark of Ian Latter.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, and mirrored at the Midnight Code web
    site; as at version 2 of the License only.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    (version 2) along with this program; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA, or see http://midnightcode.org/gplv2.txt

*/

#define TGXF_DEBUG 1

#define TGXF_CONTROL_TYPE_START 1
#define TGXF_CONTROL_TYPE_STOP 2
#define TGXF_CONTROL_TYPE_STATUS 3
#define TGXF_CONTROL_SUBTYPE_START_FILENAME 1
#define TGXF_CONTROL_SUBTYPE_START_FILESIZE 2
#define TGXF_CONTROL_SUBTYPE_START_QRCODE_VERSION 3
#define TGXF_CONTROL_SUBTYPE_START_QRCODE_FPS 4
#define TGXF_CONTROL_SUBTYPE_START_QRCODE_BYTES 5
#define TGXF_CONTROL_SUBTYPE_STOP_PAUSE 1
#define TGXF_CONTROL_SUBTYPE_STOP_COMPLETE 2
#define TGXF_CONTROL_SUBTYPE_STOP_CANCEL 3
#define TGXF_CONTROL_SUBTYPE_STATUS_SINCE 1

// TGXf Session Parameters (TGXf State Machine)
unsigned int tgxf_session_total_frames = 0;
unsigned int tgxf_session_first_data_frame = 0;
unsigned int tgxf_session_good_start = 0;
unsigned int tgxf_session_last_data_counter = 15;
unsigned int tgxf_session_frame_count = 0;
unsigned int tgxf_session_error_count = 0;
unsigned long tgxf_session_recent_crc32 = 0L;
unsigned long tgxf_session_total_crc32 = 0L;
unsigned int tgxf_session_max_errors = 5;
unsigned int tgxf_session_errors = 0;
unsigned int tgxf_session_aborted = 0;
unsigned int tgxf_session_total_bytes = 0;
unsigned int tgxf_session_complete = 0;

unsigned char tgxf_session_file_name[20];
unsigned int tgxf_session_file_size = 0;
unsigned int tgxf_session_version = 0;
unsigned int tgxf_session_qrcode_version = 0;
unsigned int tgxf_session_qrcode_fps = 0;
unsigned int tgxf_session_qrcode_bytes = 0;


unsigned int unpack_tgxf_control_payload_16bit_number(unsigned char *frame_payload, int payload_size)
{

  if(payload_size >= 2)
    return (unsigned int)frame_payload[0] << 8 | (unsigned int)frame_payload[1];

  return 0;
}


unsigned long unpack_tgxf_control_payload_32bit_number(unsigned char *frame_payload, int payload_size)
{
  if(payload_size >= 4) {
    return (unsigned long)frame_payload[0] << 24 | (unsigned long)frame_payload[1] << 16 |
      (unsigned long)frame_payload[2] << 8 | (unsigned long)frame_payload[3]; 
  }

  return 0L;
}


unsigned int unpack_tgxf_control_payload_clean_string(unsigned char *string_dest, unsigned char *frame_payload, int payload_size)
{
  unsigned char *chr_p;

  memset(string_dest, 0, 10);
  if(payload_size > 0) {
    // Store String
    if(memcpy(string_dest, frame_payload, 9) != string_dest) {
      fprintf(stderr, "ERROR(TGXf): badly formatted control string.\n");
      return -1;
    }
    string_dest[10] = '\0';
    // Clean String
    chr_p = string_dest;
    while(*chr_p) {
      // Unsavory characters are replaced
      if(*chr_p < 32 || ( *chr_p > 32 && *chr_p < 45 ) ||
        ( *chr_p > 46 && *chr_p < 48 ) || ( *chr_p > 57 && *chr_p < 65 ) ||
        ( *chr_p > 90 && *chr_p < 97 ) || *chr_p > 122) {
          *chr_p = '_';
      }
      // No leading dot
      if(*chr_p == 46 && chr_p == string_dest) {
          *chr_p = '_';
      }
      chr_p++;
    }
  }
  return 0;
}


static int decode_tgxf_frame (const zbar_symbol_t *sym)
{
  unsigned char *frame_buffer;
  unsigned char *frame_payload;
  unsigned char data_buffer[1024];
  unsigned char *data_buffer_p;
  int frame_size;
  int payload_size;
  int data_buffer_size = 1024;

  unsigned char control_byte;
  unsigned int control_bit;
  unsigned int control_type;
  unsigned int control_subtype;
  unsigned int data_counter;
  unsigned int expected_counter;
  unsigned int lost_frames;
  unsigned int transfer_bytes_remaining;
  unsigned int frames_remaining;
  unsigned int minutes_remaining;
  unsigned int seconds_remaining;
  unsigned int bytes_to_use;
  unsigned int read_len;
  FILE *fp;
  int i;
  float progress_state;

  frame_size = zbar_symbol_get_data_length(sym);
  if(frame_size <= 0) {
    fprintf(stderr, "ERROR(TGXf): invalid frame size: %d\n", frame_size);
    return -1;
  }
  if((frame_buffer = (unsigned char *)malloc(frame_size)) == NULL) {
    fprintf(stderr, "ERROR(TGXf): unable to allocate frame buffer\n");
    return -1;
  }
  if(memcpy(frame_buffer, zbar_symbol_get_data(sym), frame_size) != frame_buffer) {
    fprintf(stderr, "ERROR(TGXf): failed to establish frame buffer\n");
    return -1;
  }

  // Point to Payload
  frame_payload = frame_buffer;
  frame_payload++;
  payload_size = frame_size - 1;

  // Get Control Byte and Bit
  control_byte = *frame_buffer;
  control_bit = control_byte & 1;

  switch(control_bit) {

    // CONTROL FRAME
    case 1:
      if(TGXF_DEBUG)
        fprintf(stdout, "DEBUG(TGXf): Control Frame\n");
      control_type = (control_byte & 14) >> 1;
      // Only accept STOP if session has been aborted by client
      if(tgxf_session_aborted && control_type != TGXF_CONTROL_TYPE_STOP) {
        if(TGXF_DEBUG)
          fprintf(stdout, "DEBUG(TGXf): ignoring non-STOP conrol message in aborted transfer.\n");
        break;
      }
      control_subtype = (control_byte & 240) >> 4;
      if(TGXF_DEBUG)
        fprintf(stdout, "DEBUG(TGXf): Control Type=[%d], SubType=[%d]\n",control_type, control_subtype);
      switch(control_type) {
        case TGXF_CONTROL_TYPE_START:
          if(TGXF_DEBUG)
            fprintf(stdout, "DEBUG(TGXf): Control Type START\n");
          // Don't accept START if data frames have been received
          if(tgxf_session_first_data_frame) {
            if(TGXF_DEBUG)
              fprintf(stdout, "DEBUG(TGXf): ignoring late START conrol message in active transfer.\n");
          }
          switch(control_subtype) {
            case TGXF_CONTROL_SUBTYPE_START_FILENAME:
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): Control SubType FILENAME\n");
              unpack_tgxf_control_payload_clean_string(tgxf_session_file_name, frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/FILENAME [%s].\n", tgxf_session_file_name);
              break;
            case TGXF_CONTROL_SUBTYPE_START_FILESIZE:
              tgxf_session_file_size = unpack_tgxf_control_payload_16bit_number(frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/FILESIZE [%d].\n", tgxf_session_file_size);
              break;
            case TGXF_CONTROL_SUBTYPE_START_QRCODE_VERSION:
              tgxf_session_qrcode_version = unpack_tgxf_control_payload_16bit_number(frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/QRCODE_VERSION [%d].\n", tgxf_session_qrcode_version);
              break;
            case TGXF_CONTROL_SUBTYPE_START_QRCODE_FPS:
              tgxf_session_qrcode_fps = unpack_tgxf_control_payload_16bit_number(frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/QRCODE_FPS [%d].\n", tgxf_session_qrcode_fps);
              break;
            case TGXF_CONTROL_SUBTYPE_START_QRCODE_BYTES:
              tgxf_session_qrcode_bytes = unpack_tgxf_control_payload_16bit_number(frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/QRCODE_BYTES [%d].\n", tgxf_session_qrcode_bytes);
              break;
            default:
              fprintf(stdout, "WARNING(TGXf): reserved control sub-type used, frame ignored.\n");
              return -1;
          }
          break;

        case TGXF_CONTROL_TYPE_STOP:
          if(TGXF_DEBUG)
            fprintf(stdout, "DEBUG(TGXf): Control Type STOP\n");
          switch(control_subtype) {
            case TGXF_CONTROL_SUBTYPE_STOP_PAUSE:
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): STOP/PAUSE.\n");
              break;
            case TGXF_CONTROL_SUBTYPE_STOP_COMPLETE:
              tgxf_session_total_crc32 = unpack_tgxf_control_payload_32bit_number(frame_payload, payload_size);
              if(!tgxf_session_good_start) {
                if(TGXF_DEBUG)
                  fprintf(stdout, "WARNING(TGXf): failed to get required start data, session aborted.\n");
                tgxf_session_aborted = 1;
                break;
              }
              if((fp = fopen((const char *)tgxf_session_file_name, "rb")) == NULL) {
                fprintf(stdout, "ERROR(TGXf): unable to open file for reading, at %s.\n", tgxf_session_file_name);
                break;
              }
              tgxf_session_recent_crc32 = 0L;
              data_buffer_p = data_buffer;
              while((read_len = fread(data_buffer, 1, data_buffer_size, fp)) > 0) {
                tgxf_session_recent_crc32 = update_crc(tgxf_session_recent_crc32, data_buffer_p, read_len);
              }
              if(tgxf_session_total_crc32 != tgxf_session_recent_crc32) {
                fprintf(stdout, "ERROR(TGXf): CRC doesn't match, calculated %lu, received %lu.\n",
                  tgxf_session_recent_crc32, tgxf_session_total_crc32);
              } else {
                tgxf_session_complete = 1;
                fprintf(stdout, "(TGXf): CRC validated, transfer successful.\n");
              }
              // if(TGXF_DEBUG)
                // fprintf(stdout, "DEBUG(TGXf): STOP/COMPLETE [%u].\n", tgxf_session_total_crc32);
              // SESSION RESET
              break;
            case TGXF_CONTROL_SUBTYPE_STOP_CANCEL:
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): STOP/CANCEL.\n");
              // SESSION RESET
              break;
            default:
              fprintf(stdout, "WARNING(TGXf): reserved control sub-type used, frame ignored.\n");
              return -1;
          }
          break;

        case TGXF_CONTROL_TYPE_STATUS:
          if(TGXF_DEBUG)
            fprintf(stdout, "DEBUG(TGXf): Control Type STATUS\n");
          switch(control_subtype) {
            case TGXF_CONTROL_SUBTYPE_STATUS_SINCE:
              tgxf_session_recent_crc32 = unpack_tgxf_control_payload_32bit_number(frame_payload, payload_size);
              // if(TGXF_DEBUG)
                // fprintf(stdout, "DEBUG(TGXf): STATUS/SINCE [%u].\n", tgxf_session_recent_crc32);
              break;
            default:
              fprintf(stdout, "WARNING(TGXf): reserved control sub-type used, frame ignored.\n");
              return -1;
          }
          break;

        default:
          fprintf(stdout, "WARNING(TGXf): reserved control type used, frame ignored.\n");
          return -1;
      }
      break;

    // DATA FRAME
    case 0:
      if(TGXF_DEBUG)
        fprintf(stdout, "DEBUG(TGXf): Data Frame ------------------------------\n");

      // Ignore data frames if session aborted
      if(tgxf_session_aborted) {
        if(TGXF_DEBUG)
          fprintf(stdout, "WARNING(TGXf): session aborted, ignoring data frames.\n");
        break;
      }

      // If this is the first Data Frame then do we have the right START info to continue?
      if(!tgxf_session_first_data_frame) {
        tgxf_session_first_data_frame = 1;
        fprintf(stdout, "-= TGXf =-\n");
        fprintf(stdout, "  Filename:                 %s\n", tgxf_session_file_name);
        fprintf(stdout, "  File size:                %d\n", tgxf_session_file_size);
        fprintf(stdout, "  QRcode Version:           %d\n", tgxf_session_qrcode_version);
        fprintf(stdout, "  QRcode Frames per Second: %d\n", tgxf_session_qrcode_fps);
        fprintf(stdout, "  QRcode Bytes per Frame:   %d\n", tgxf_session_qrcode_bytes);
        fprintf(stdout, "\n");
        // Do we have sufficient START information to perform the data transfer?
        if(strlen((const char *)tgxf_session_file_name) > 0 &&
          tgxf_session_file_size > 0 &&
          tgxf_session_qrcode_fps > 0 &&
          tgxf_session_qrcode_bytes > 0) {
            tgxf_session_good_start = 1;
            tgxf_session_total_frames = tgxf_session_file_size / tgxf_session_qrcode_bytes;
            if(tgxf_session_total_frames * tgxf_session_qrcode_bytes < tgxf_session_file_size)
              tgxf_session_total_frames++;
        }
      }
      if(!tgxf_session_good_start) {
        if(TGXF_DEBUG)
          fprintf(stdout, "WARNING(TGXf): failed to get required start data, session aborted.\n");
        tgxf_session_aborted = 1;
        break;
      }

      // Duplicate Data Frame Detection - Ignore
      data_counter = (control_byte & 30) >> 1;
      if(data_counter == tgxf_session_last_data_counter) {
        if(TGXF_DEBUG)
          fprintf(stdout, "WARNING(TGXf): duplicate data frame, frame ignored.\n");
        break;
      }

      // Missing Data Frame Detection - Accumulate Errors
      expected_counter = tgxf_session_last_data_counter + 1;
      if(expected_counter == 16)
        expected_counter = 0;
      if(TGXF_DEBUG)
        fprintf(stdout, "DEBUG(TGXf): data_counter=%02d expected_counter=%02d; frame %d.\n",
          data_counter, expected_counter, tgxf_session_frame_count);
      if(data_counter != expected_counter) {
        lost_frames = data_counter - expected_counter;
        if(expected_counter > data_counter) {
          lost_frames = data_counter + 16 - expected_counter;
        }
        if(lost_frames == 1) {
          fprintf(stdout, "WARNING(TGXf): lost 1 data frame, error recorded.\n");
        } else {
          fprintf(stdout, "WARNING(TGXf): lost %d data frames, error recorded.\n", lost_frames);
        }
        // Error Accumulation
        tgxf_session_errors += lost_frames;
        if(tgxf_session_errors >= tgxf_session_max_errors) {
          tgxf_session_aborted = 1;
          fprintf(stdout, "ERROR(TGXf): maximum errors for session reached, transfer aborted.\n");
        }
        // Correct the session frame count 
        tgxf_session_frame_count += lost_frames;
      }

      // Do last-packet calculation and under-run detection
      transfer_bytes_remaining = 
        tgxf_session_file_size - (tgxf_session_qrcode_bytes * tgxf_session_frame_count);
      bytes_to_use = tgxf_session_qrcode_bytes;
      if(transfer_bytes_remaining < tgxf_session_qrcode_bytes)
        bytes_to_use = transfer_bytes_remaining;
      tgxf_session_total_bytes += bytes_to_use;
      if(TGXF_DEBUG) {
        fprintf(stdout, "DEBUG(TGXf): payload_size=%d bytes_to_use=%d qrcode_bytes=%d; frame %d.\n", payload_size, bytes_to_use, tgxf_session_qrcode_bytes, tgxf_session_frame_count);
        fprintf(stdout, "DEBUG(TGXf): session_total_bytes after this frame is written = %d/%d.\n", tgxf_session_total_bytes, tgxf_session_file_size);
      }
      if(payload_size < bytes_to_use) {
        fprintf(stdout, "ERROR(TGXf): undersized payload on frame %d.\n", tgxf_session_frame_count);
        break;
      }

      // Update Display
      frames_remaining = tgxf_session_total_frames - tgxf_session_frame_count;
      seconds_remaining = frames_remaining / tgxf_session_qrcode_fps;
      minutes_remaining = seconds_remaining / 60;
      seconds_remaining = seconds_remaining % 60;
      progress_state = (float)20 - ((float)frames_remaining / (float)tgxf_session_total_frames * (float)20);
      fprintf(stdout, "(TGXf): Frame: %4d/%d ",
        tgxf_session_frame_count + 1, tgxf_session_total_frames);
      fprintf(stdout, " Errors: %2d/%d ",
        tgxf_session_errors, tgxf_session_max_errors);
      fprintf(stdout, " [");
      for(i = 0; i < 20; i++ ) {
        if(i >= progress_state) {
          fprintf(stdout, "%c", ' ' );
        } else {
          fprintf(stdout, "%c", '=' );
        }
      }
      fprintf(stdout, "] %02d min %02d sec to go.\r",
        minutes_remaining, seconds_remaining);
      if(TGXF_DEBUG || frames_remaining == 1)
        fprintf(stdout, "\n");

      // Write Data to File
      if(access((const char *)tgxf_session_file_name, F_OK) == -1) {
        if((fp = fopen((const char *)tgxf_session_file_name, "w+b")) == NULL) {
          fprintf(stdout, "ERROR(TGXf): unable to create/open file for reading/writing, at %s.\n", tgxf_session_file_name);
          break;
        }
      } else {
        if((fp = fopen((const char *)tgxf_session_file_name, "r+b")) == NULL) {
          fprintf(stdout, "ERROR(TGXf): unable to open file for reading/writing, at %s.\n", tgxf_session_file_name);
          break;
        }
        // TODO: need to check on seek past EOF
        if(fseek(fp, (tgxf_session_qrcode_bytes * tgxf_session_frame_count), SEEK_SET) < 0) {
          fprintf(stdout, "ERROR(TGXf): unable to seek file, at %s.\n", tgxf_session_file_name);
          fclose(fp);
          break;
        }
      }
      if(fwrite((const void *)frame_payload, (size_t)bytes_to_use, (size_t)1, fp) != (size_t)1) {
        fprintf(stdout, "ERROR(TGXf): unable to write to file, at %s.\n", tgxf_session_file_name);
        fclose(fp);
        break;
      }
      fclose(fp);

      // Increment tally
      tgxf_session_last_data_counter = data_counter;
      tgxf_session_frame_count++;
      break;

  }

  return(1);
}

static void data_handler (zbar_image_t *img, const void *userdata)
{
    const zbar_symbol_t *sym = zbar_image_first_symbol(img);
    assert(sym);
    int n = 0;
    for(; sym; sym = zbar_symbol_next(sym)) {
        if(zbar_symbol_get_count(sym))
            continue;

        zbar_symbol_type_t type = zbar_symbol_get_type(sym);
        if(type == ZBAR_PARTIAL)
            continue;

        if(!format) {
            printf("%s:", zbar_get_symbol_name(type));
            if(fwrite(zbar_symbol_get_data(sym),
                      zbar_symbol_get_data_length(sym),
                      1, stdout) != 1)
                continue;
        }
        else if(format == RAW) {
            if(fwrite(zbar_symbol_get_data(sym),
                      zbar_symbol_get_data_length(sym),
                      1, stdout) != 1)
                continue;
        }
        else if(format == TGXF) {
            decode_tgxf_frame(sym);
            continue;
        }
        else if(format == XML) {
            if(!n)
                printf("<index num='%u'>\n", zbar_image_get_sequence(img));
            zbar_symbol_xml(sym, &xml_buf, &xml_len);
            if(fwrite(xml_buf, xml_len, 1, stdout) != 1)
                continue;
        }
        printf("\n");
        n++;
    }

    if(format == XML && n)
        printf("</index>\n");
    fflush(stdout);

    if(!quiet && n)
        fprintf(stderr, BELL);
}

int main (int argc, const char *argv[])
{
    /* setup zbar library standalone processor,
     * threads will be used if available
     */
    proc = zbar_processor_create(1);
    if(!proc) {
        fprintf(stderr, "ERROR: unable to allocate memory?\n");
        return(1);
    }
    zbar_processor_set_data_handler(proc, data_handler, NULL);

    const char *video_device = "";
    int display = 1;
    unsigned long infmt = 0, outfmt = 0;
    int i;
    for(i = 1; i < argc; i++) {
        if(argv[i][0] != '-')
            video_device = argv[i];
        else if(argv[i][1] != '-') {
            int j;
            for(j = 1; argv[i][j]; j++) {
                if(argv[i][j] == 'S') {
                    if(!argv[i][++j]) {
                        i++;
                        j = 0;
                    }
                    if(parse_config(&argv[i][j], i, argc, "-S"))
                        return(usage(1));
                    break;
                }
                switch(argv[i][j]) {
                case 'h': return(usage(0));
                case 'v': zbar_increase_verbosity(); break;
                case 'q': quiet = 1; break;
                default:
                    fprintf(stderr, "ERROR: unknown bundled config: -%c\n\n",
                            argv[i][j]);
                    return(usage(1));
                }
            }
        }
        else if(!argv[i][2]) {
            if(i < argc - 1)
                video_device = argv[argc - 1];
            break;
        }
        else if(!strcmp(argv[i], "--help"))
            return(usage(0));
        else if(!strcmp(argv[i], "--version"))
            return(printf(PACKAGE_VERSION "\n") <= 0);
        else if(!strcmp(argv[i], "--set")) {
            i++;
            if(parse_config(argv[i], i, argc, "--set"))
                return(usage(1));
        }
        else if(!strncmp(argv[i], "--set=", 6)) {
            if(parse_config(&argv[i][6], i, argc, "--set="))
                return(usage(1));
        }
        else if(!strcmp(argv[i], "--quiet"))
            quiet = 1;
        else if(!strcmp(argv[i], "--xml"))
            format = XML;
        else if(!strcmp(argv[i], "--raw"))
            format = RAW;
        else if(!strcmp(argv[i], "--TGXf"))
            format = TGXF;
        else if(!strcmp(argv[i], "--nodisplay"))
            display = 0;
        else if(!strcmp(argv[i], "--verbose"))
            zbar_increase_verbosity();
        else if(!strncmp(argv[i], "--verbose=", 10))
            zbar_set_verbosity(strtol(argv[i] + 10, NULL, 0));
        else if(!strncmp(argv[i], "--prescale=", 11)) {
            char *x = NULL;
            long int w = strtol(argv[i] + 11, &x, 10);
            long int h = 0;
            if(x && *x == 'x')
                h = strtol(x + 1, NULL, 10);
            if(!w || !h || !x || *x != 'x') {
                fprintf(stderr, "ERROR: invalid prescale: %s\n\n", argv[i]);
                return(usage(1));
            }
            zbar_processor_request_size(proc, w, h);
        }
        else if(!strncmp(argv[i], "--v4l=", 6)) {
            long int v = strtol(argv[i] + 6, NULL, 0);
            zbar_processor_request_interface(proc, v);
        }
        else if(!strncmp(argv[i], "--iomode=", 9)) {
            long int v = strtol(argv[i] + 9, NULL, 0);
            zbar_processor_request_iomode(proc, v);
        }
        else if(!strncmp(argv[i], "--infmt=", 8) &&
                strlen(argv[i]) == 12)
            infmt = (argv[i][8] | (argv[i][9] << 8) |
                     (argv[i][10] << 16) | (argv[i][11] << 24));
        else if(!strncmp(argv[i], "--outfmt=", 9) &&
                strlen(argv[i]) == 13)
            outfmt = (argv[i][9] | (argv[i][10] << 8) |
                      (argv[i][11] << 16) | (argv[i][12] << 24));
        else {
            fprintf(stderr, "ERROR: unknown option argument: %s\n\n",
                    argv[i]);
            return(usage(1));
        }
    }

    if(infmt || outfmt)
        zbar_processor_force_format(proc, infmt, outfmt);

    /* open video device, open window */
    if(zbar_processor_init(proc, video_device, display) ||
       /* show window */
       (display && zbar_processor_set_visible(proc, 1)))
        return(zbar_processor_error_spew(proc, 0));

    if(format == XML) {
#ifdef _WIN32
        fflush(stdout);
        _setmode(_fileno(stdout), _O_BINARY);
#endif
        printf(xml_head, video_device);
        fflush(stdout);
    }

    /* start video */
    int active = 1;
    if(zbar_processor_set_active(proc, active))
        return(zbar_processor_error_spew(proc, 0));

    /* let the callback handle data */
    int rc;
    while((rc = zbar_processor_user_wait(proc, -1)) >= 0) {
        if(tgxf_session_complete) {
            printf("ThruGlassXfer session completed successfully.\n");
            break;
        }
        if(tgxf_session_aborted) {
            printf("ThruGlassXfer session aborted.\n");
            break;
        }
        if(rc == 'q' || rc == 'Q')
            break;
        if(rc == ' ') {
            active = !active;
            if(zbar_processor_set_active(proc, active))
                return(zbar_processor_error_spew(proc, 0));
        }
    }

    /* report any errors that aren't "window closed" */
    if(rc && rc != 'q' && rc != 'Q' &&
       zbar_processor_get_error_code(proc) != ZBAR_ERR_CLOSED)
        return(zbar_processor_error_spew(proc, 0));

    /* free resources (leak check) */
    zbar_processor_destroy(proc);

    if(format == XML) {
        printf("%s", xml_foot);
        fflush(stdout);
    }
    return(0);
}
